from pathlib import Path
from typing import Optional, Union

from pycosmiconfig.port.loaders import load_toml
from pycosmiconfig.port.types import CosmiconfigResult, Config
from pycosmiconfig.port.ExplorerBase import ExplorerBase, get_extension_description
from pycosmiconfig.port.util import emplace, get_property_by_path


class ExplorerSync(ExplorerBase):
    """@internal"""

    def load(self, filepath: str) -> CosmiconfigResult:
        filepath = Path(filepath).resolve()

        def load() -> CosmiconfigResult:
            return self._config.transform(self._read_configuration(filepath))

        if self._load_cache:
            return emplace(self._load_cache, str(filepath), load)
        return load()

    def search(self, from_dir_str: str = "") -> Optional[CosmiconfigResult]:
        if self._config.meta_config_file_path:
            self._loading_meta_config = True
            config = self.load(self._config.meta_config_file_path)
            self._loading_meta_config = False
            if config and not config.is_empty:
                return config

        stop_dir = Path(self._config.stop_dir).resolve()
        from_dir = Path(from_dir_str).resolve()

        def _search() -> Optional[CosmiconfigResult]:
            nonlocal from_dir
            if from_dir.is_dir():
                for place in self._config.search_places:
                    filepath = from_dir / place
                    try:
                        result = self._read_configuration(filepath)
                        if result and not (
                            result.is_empty and self._config.ignore_empty_search_places
                        ):
                            return self._config.transform(result)
                    except FileNotFoundError:
                        continue
                    except IsADirectoryError:
                        continue
                    except NotADirectoryError:
                        continue

            parent_dir = from_dir.parent.resolve()
            if from_dir != stop_dir and from_dir != parent_dir:
                from_dir = parent_dir
                if self._search_cache is not None:
                    return emplace(self._search_cache, str(from_dir), _search)
                return _search()
            return self._config.transform(None)

        if self._search_cache is not None:
            return emplace(self._search_cache, str(from_dir), _search)
        return _search()

    def _read_configuration(self, filepath: Union[str, Path]) -> CosmiconfigResult:
        filepath = Path(filepath)
        with open(str(filepath), 'r') as f:
            contents = f.read()
        config = self._load_configuration(filepath, contents)
        return self._to_cosmiconfig_result(str(filepath), config)

    def _load_configuration(self, filepath: Path, contents: str) -> Optional[Config]:
        if not contents.strip():
            return None

        # PORT COMMENT: JS version reads `package.json`
        if filepath.name in ["pyproject.toml", "pyproject.tml"]:
            pyproject_conf = load_toml(str(filepath), contents)
            return get_property_by_path(pyproject_conf, self._config.package_prop)

        extension = filepath.suffix
        try:
            loader = self._config.loaders.get(
                extension or "noExt", self._config.loaders.get("default", None)
            )
            if loader:
                return loader(str(filepath), contents)
        except Exception as error:
            error.filepath = str(filepath)
            raise error
        raise ValueError(
            f"No loader specified for {get_extension_description(extension)}"
        )



from pycosmiconfig.port.util import get_property_by_path

source = {
    "ant": {
        "beetle": {
            "cootie": {
                "flea": "foo",
            },
            "louse": {
                "vermin": "bar",
            },
        },
        "fancy.name": {
            "another.fancy.name": 9,
        },
    },
    "ant.beetle.cootie": 333,
}


def describe_get_property_by_path():
    def test_property_name_includes_period():
        assert get_property_by_path(source, "ant.beetle.cootie") == 333

    def describe_period_delimited_string_path():
        def test_returns_defined_value():
            assert get_property_by_path(source, "ant") == source["ant"]
            assert get_property_by_path(source, "ant.beetle.cootie.flea") == "foo"
            assert (
                get_property_by_path(source, "ant.beetle.louse")
                == source["ant"]["beetle"]["louse"]
            )

        def test_returns_undefined():
            assert get_property_by_path(source, "beetle") is None
            assert get_property_by_path(source, "ant.beetle.cootie.fleeee") is None
            assert get_property_by_path(source, "ant.beetle.vermin") is None
            assert get_property_by_path(source, "ant.fancy.name") is None

    def describe_array_path():
        def test_returns_defined_value():
            assert get_property_by_path(source, ["ant"]) == source["ant"]
            assert (
                get_property_by_path(source, ["ant", "beetle", "cootie", "flea"])
                == "foo"
            )
            assert (
                get_property_by_path(source, ["ant", "beetle", "louse"])
                == source["ant"]["beetle"]["louse"]
            )

        def test_returns_undefined():
            assert get_property_by_path(source, ["beetle"]) is None
            assert (
                get_property_by_path(source, ["ant", "beetle", "cootie", "fleeee"])
                is None
            )
            assert get_property_by_path(source, ["ant", "beetle", "vermin"]) is None

        def test_handles_property_names_with_periods():
            assert (
                get_property_by_path(
                    source, ["ant", "fancy.name", "another.fancy.name"]
                )
                == 9
            )
            assert (
                get_property_by_path(
                    source, ["ant", "fancy.name", "another.fancy.name", "foo"]
                )
                is None
            )
            assert get_property_by_path(source, ["ant", "fancy.namez"]) is None



import os
from pathlib import Path
import shutil
import tempfile
from unittest.mock import Mock
from typing import List


def normalize_directory_slash(pathname: str) -> str:
    return pathname.replace("\\", "/")


class TempDir:
    def __init__(self):
        # Get the actual path for temp directories that are symlinks (MacOS).
        # Without the actual path, tests that use process.chdir will unexpectedly
        # return the real path instead of symlink path.
        temp_dir = os.path.realpath(tempfile.gettempdir())

        # Get the pathname of the file that imported util.py.
        # Used to create a unique directory name for each test suite.
        parent = __file__ or "cosmiconfig"
        relative_parent = Path(parent).relative_to(Path.cwd())

        # Each temp directory will be unique to the test file.
        # This ensures that temp files/dirs won't cause side effects for other tests.
        self.dir = Path(temp_dir, "cosmiconfig", f"{relative_parent}-dir")
        self.dir.mkdir(parents=True, exist_ok=True)

    def absolute_path(self, dir_: str) -> Path:
        # Join paths to ensure dir is always inside the working temp directory
        return self.dir / dir_

    def create_dir(self, dir_: str) -> None:
        dirname = self.absolute_path(dir_)
        dirname.mkdir(parents=True, exist_ok=True)

    def create_file(self, file: str, contents: str) -> None:
        file_path = self.absolute_path(file)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_text(f"{contents}\n")

    def get_spy_path_calls(self, spy: Mock) -> List[str]:
        calls = spy.call_args_list
        result = []
        for call in calls:
            file_path = call.args[0]
            relative_path = Path(file_path).relative_to(self.dir)
            normalize_cross_platform = normalize_directory_slash(str(relative_path))
            result.append(normalize_cross_platform)
        return result

    def clean(self) -> None:
        shutil.rmtree(self.dir, ignore_errors=True)

    def delete_temp_dir(self) -> None:
        shutil.rmtree(self.dir, ignore_errors=True)


def is_not_mjs(file_path: str) -> bool:
    return Path(file_path).suffix != ".mjs"



from pycosmiconfig.port.types import (
    Config,
    CosmiconfigResult,
    LoaderResult,
    Loader,
    Loaders,
    Transform,
    Options,
    PublicExplorer,
    PublicExplorerLoadFn,
    PublicExplorerSearchFn,
)
from pycosmiconfig.port.loaders import load_py, load_json, load_toml, load_yaml
from pycosmiconfig.port.index import meta_search_places, default_loaders, cosmiconfig



class imdict(dict):
    '''
    Simple immutable dict. Don't expect miracles.

    See https://stackoverflow.com/a/11026765/9788634
    '''

    def __hash__(self):
        return id(self)

    def _immutable(self, *args, **kws):
        raise TypeError("object is immutable")

    __setitem__ = _immutable
    __delitem__ = _immutable
    clear = _immutable
    update = _immutable
    setdefault = _immutable
    pop = _immutable
    popitem = _immutable



import builtins
import os
import pytest
from unittest.mock import patch

from pycosmiconfig import cosmiconfig, Options, CosmiconfigResult
from util import TempDir


@pytest.fixture(autouse=True)
def temp():
    temp_dir = TempDir()
    yield temp_dir
    temp_dir.delete_temp_dir()


def test_throws_when_supplied_loaders(temp: TempDir):
    temp.create_file(".config.yml", "pycosmiconfig:\n  loaders: []")

    current_dir = os.getcwd()
    os.chdir(temp.dir)

    with pytest.raises(ValueError):
        cosmiconfig("foo")

    os.chdir(current_dir)


def describe_uses_user_configured_search_places_without_placeholders():
    @pytest.fixture(autouse=True)
    def current_dir(temp: TempDir):
        temp.create_dir("sub")
        temp.create_file(".foorc", "a: b")
        temp.create_file(".foo-config", "a: c")

        current_dir = os.getcwd()
        os.chdir(temp.dir)
        yield current_dir
        os.chdir(current_dir)

    def _run_test(temp: TempDir):
        file = temp.absolute_path(".foo-config")
        start_dir = temp.absolute_path("sub")
        explorer_options = Options(stop_dir=str(temp.absolute_path(".")))

        with patch.object(builtins, "open", wraps=builtins.open) as open_spy:
            # Check read files as cosmiconfig searches for meta config
            explorer = cosmiconfig("foo", **explorer_options.dict())
            meta_config_search_paths = temp.get_spy_path_calls(open_spy)
            assert meta_config_search_paths == [
                "pyproject.toml",
                "pyproject.tml",
                ".config.json",
                ".config.yaml",
                ".config.yml",
            ]
            open_spy.reset_mock()

            # Check read files as cosmiconfig searches for package config
            result = explorer.search(start_dir)
            config_search_paths = temp.get_spy_path_calls(open_spy)
            assert config_search_paths == [
                ".config.yml",
                "sub/.foo-config",
                "sub/.foo.config.yml",
                ".foo-config",
            ]
            assert result == CosmiconfigResult(config={"a": "c"}, filepath=str(file))

    def test_without_placeholder(temp: TempDir):
        temp.create_file(
            ".config.yml",
            'pycosmiconfig:\n  search_places: [".foo-config", ".foo.config.yml"]',
        )
        _run_test(temp)

    def test_with_placeholder(temp: TempDir):
        temp.create_file(
            ".config.yml",
            'pycosmiconfig:\n  search_places: [".{name}-config", ".{name}.config.yml"]',
        )
        _run_test(temp)


def describe_checks_config_in_meta_file():
    @pytest.fixture(autouse=True)
    def current_dir(temp: TempDir):
        temp.create_dir("sub")
        temp.create_file(".foo-config", "a: c")

        current_dir = os.getcwd()
        os.chdir(temp.dir)
        yield current_dir
        os.chdir(current_dir)

    def describe_not_existing():
        def test_sync(temp: TempDir):
            temp.create_file(
                ".config.yml", 'pycosmiconfig:\n  search_places: [".foo-config"]'
            )
            file = temp.absolute_path(".foo-config")
            explorer_options = Options(
                stop_dir=str(temp.absolute_path(".")), ignore_empty_search_places=False
            )

            with patch.object(builtins, "open", wraps=builtins.open) as open_spy:
                # Check read files as cosmiconfig searches for meta config
                explorer = cosmiconfig("foo", **explorer_options.dict())
                meta_config_search_paths = temp.get_spy_path_calls(open_spy)

                assert meta_config_search_paths == [
                    "pyproject.toml",
                    "pyproject.tml",
                    ".config.json",
                    ".config.yaml",
                    ".config.yml",
                ]
                open_spy.reset_mock()

                # Check read files as cosmiconfig searches for package config
                result = explorer.search(str(temp.dir))

                config_search_paths = temp.get_spy_path_calls(open_spy)
                assert config_search_paths == [
                    ".config.yml",
                    ".foo-config",
                ]
                assert result == CosmiconfigResult(
                    config={"a": "c"}, filepath=str(file)
                )

    def describe_existing():
        def test_sync(temp: TempDir):
            temp.create_file(".config.yml", "foo:\n  a: d")
            explorer_options = Options(stop_dir=str(temp.absolute_path(".")))

            explorer = cosmiconfig("foo", **explorer_options.dict())
            # Check read files as cosmiconfig searches for package config
            with patch.object(builtins, "open", wraps=builtins.open) as open_spy:
                result = explorer.search(str(temp.dir))

                config_search_paths = temp.get_spy_path_calls(open_spy)
                assert config_search_paths == [".config.yml"]
                assert result == CosmiconfigResult(
                    config={"a": "d"}, filepath=str(temp.absolute_path(".config.yml"))
                )



from typing import Any, Dict, List, Union, TypeVar, Callable, Optional

K = TypeVar("K")
V = TypeVar("V")


def emplace(cache: Dict[K, V], key: K, fn: Callable[[], V]) -> V:
    """@internal"""
    cached = cache.get(key, None)
    if cached is not None:
        return cached
    result = fn()
    cache[key] = result
    return result


def get_property_by_path(source: Dict[str, Any], path: Union[str, List[str]]) -> Any:
    """
    @internal

    Resolves property names or property paths defined with period-delimited
    strings or arrays of strings. Property names that are found on the source
    object are used directly (even if they include a period).

    Nested property names that include periods, within a path, are only
    understood in array paths.
    """
    if isinstance(path, str) and path in source:
        return source[path]

    parsed_path = path.split(".") if isinstance(path, str) else path
    previous = source
    for key in parsed_path:
        if previous is None:
            break
        try:
            previous = previous.get(key, None)
        except AttributeError:
            previous = None
    return previous


def remove_none_values_from_object(
    options: Optional[Dict[str, Any]]
) -> Optional[Dict[str, Any]]:
    """@internal"""
    if options is None:
        return None
    return {key: value for key, value in options.items() if value is not None}



from pathlib import Path
from typing import Optional

from pycosmiconfig.port.types import InternalOptions, Cache, Config, CosmiconfigResult
from pycosmiconfig.port.util import get_property_by_path


class ExplorerBase:
    """@internal"""

    def __init__(self, options: InternalOptions) -> None:
        self._loading_meta_config: bool = False
        self._config: InternalOptions = options
        self._load_cache: Optional[Cache] = None
        self._search_cache: Optional[Cache] = None

        if options.cache:
            self._load_cache = {}
            self._search_cache = {}

        self._validate_config()

    def _validate_config(self) -> None:
        for place in self._config.search_places:
            extension = Path(place).suffix
            loader = self._config.loaders.get(extension or "noExt", None)
            if loader is None:
                loader = self._config.loaders.get("default", None)

            if loader is None:
                raise ValueError(
                    f"Missing loader for {get_extension_description(extension)}."
                )
            if not callable(loader):
                desc = get_extension_description(extension)
                loader_type = type(loader).__name__
                raise TypeError(
                    f"Loader for {desc} is not a function: Received {loader_type}."
                )

    def clear_load_cache(self) -> None:
        if self._load_cache:
            self._load_cache.clear()

    def clear_search_cache(self) -> None:
        if self._search_cache:
            self._search_cache.clear()

    def clear_caches(self) -> None:
        self.clear_load_cache()
        self.clear_search_cache()

    def _to_cosmiconfig_result(
        self, filepath: str, config: Config
    ) -> CosmiconfigResult:
        # PORT COMMENT: We don't distinguish between null and undefined
        if config is None:
            return CosmiconfigResult(filepath=filepath, config=None, is_empty=True)

        if (
            self._config.apply_package_property_path_to_configuration
            or self._loading_meta_config
        ):
            prop_name = self._config.config_prop
            if Path(filepath).name in ["pyproject.toml", "pyproject.tml"]:
                prop_name = self._config.package_prop
            config = get_property_by_path(config, prop_name)
        if config is None:
            return CosmiconfigResult(filepath=filepath, config=None, is_empty=True)

        return CosmiconfigResult(config=config, filepath=filepath)


def get_extension_description(extension: Optional[str]) -> str:
    """@internal"""
    return f'extension "{extension}"' if extension else "files without extensions"



import os
from typing import Optional, Union, List

from pycosmiconfig.utils import imdict
from pycosmiconfig.port.types import Options, InternalOptions
from pycosmiconfig.port.util import remove_none_values_from_object
from pycosmiconfig.port.loaders import load_py, load_json, load_toml, load_yaml
from pycosmiconfig.port.ExplorerSync import ExplorerSync
from pycosmiconfig.port.types import Loaders, Transform


# this needs to be hardcoded, as this is intended for end users,
# who can't supply options at this point
meta_search_places = [
    # PORT COMMENT: use `pyproject` instead of `package.json`
    "pyproject.toml",
    "pyproject.tml",
    ".config.json",
    ".config.yaml",
    ".config.yml",
    ".config.toml",
    ".config.tml",
    ".config.py",
]

# do not allow mutation of default loaders. Make sure it is set inside options
default_loaders = imdict(
    {
        ".py": load_py,
        ".json": load_json,
        ".yaml": load_yaml,
        ".yml": load_yaml,
        ".toml": load_toml,
        ".tml": load_toml,
        "noExt": load_yaml,
    }
)

# PORT COMMENT: No async loaders at the moment


def _identity(x):
    return x


def _get_internal_options(module_name: str, options: Options):
    InternalOptions.update_forward_refs()
    meta_explorer = ExplorerSync(
        InternalOptions(
            package_prop="tool.pycosmiconfig",
            config_prop="pycosmiconfig",
            stop_dir=os.getcwd(),
            search_places=meta_search_places,
            ignore_empty_search_places=False,
            apply_package_property_path_to_configuration=True,
            loaders=default_loaders,
            transform=_identity,
            cache=True,
            meta_config_file_path=None,
        )
    )
    meta_config_result = meta_explorer.search()

    if not meta_config_result:
        return InternalOptions(**options.dict())

    meta_config = meta_config_result.config or {}
    if "loaders" in meta_config:
        raise ValueError("Can not specify loaders in meta config file")

    override_options = InternalOptions(**meta_config)

    if override_options.search_places:
        override_options.search_places = [
            path.replace("{name}", module_name)
            for path in override_options.search_places
        ]

    override_options.meta_config_file_path = meta_config_result.filepath

    merged_options = {**options.dict()}
    merged_options.update(remove_none_values_from_object(override_options.dict()))
    return InternalOptions(**merged_options)


def _normalize_options(module_name: str, options: InternalOptions):
    defaults = InternalOptions(
        package_prop=f"tool.{module_name}",
        config_prop=module_name,
        search_places=[
            # PORT COMMENT: use `pyproject` instead of `package.json`
            "pyproject.toml",
            "pyproject.tml",
            f".{module_name}rc",
            f".{module_name}rc.json",
            f".{module_name}rc.yaml",
            f".{module_name}rc.yml",
            f".{module_name}rc.toml",
            f".{module_name}rc.tml",
            f".{module_name}rc.py",
            f".config/{module_name}rc",
            f".config/{module_name}rc.json",
            f".config/{module_name}rc.yaml",
            f".config/{module_name}rc.yml",
            f".config/{module_name}rc.toml",
            f".config/{module_name}rc.tml",
            f".config/{module_name}rc.py",
            f"{module_name}.config.py",
        ],
        ignore_empty_search_places=True,
        stop_dir=os.path.expanduser("~"),
        cache=True,
        transform=_identity,
        loaders=default_loaders,
        meta_config_file_path=None,
    )

    merged_loaders = defaults.loaders or {}
    merged_loaders.update(options.loaders or {})

    merged_options = {**defaults.dict()}
    merged_options.update(remove_none_values_from_object(options.dict()))
    merged_options.update({"loaders": merged_loaders})

    return InternalOptions(**merged_options)


# PORT COMMENT: No async at the moment
def cosmiconfig(
    module_name: str,
    package_prop: Optional[Union[str, List[str]]] = None,
    search_places: Optional[List[str]] = None,
    ignore_empty_search_places: Optional[bool] = None,
    stop_dir: Optional[str] = None,
    cache: Optional[bool] = None,
    loaders: Optional[Loaders] = None,
    transform: Optional[Transform] = None,
):
    # PORT COMMENT: Handle the difference between prop in pyproject.toml
    #               and elsewhere
    config_prop = package_prop
    package_prop = f'tool.{package_prop}'

    Options.update_forward_refs()
    options = Options(
        package_prop=package_prop,
        config_prop=config_prop,
        search_places=search_places,
        ignore_empty_search_places=ignore_empty_search_places,
        stop_dir=stop_dir,
        cache=cache,
        loaders=loaders,
        transform=transform,
    )

    internal_options = _get_internal_options(module_name, options)
    normalized_options = _normalize_options(module_name, internal_options)
    explorer = ExplorerSync(normalized_options)
    return explorer



from typing import (
    Optional,
    Callable,
    Union,
    List,
    Dict,
    Any,
    Protocol,
    runtime_checkable,
)

from pydantic import BaseModel, Field


Config = Any
"""@public"""


class CosmiconfigResult(BaseModel):
    """@public"""

    config: Config
    filepath: str
    is_empty: Optional[bool] = Field(default=None)


LoaderResult = Optional[Config]
"""@public"""


# PORT COMMENT: No async at the moment
@runtime_checkable
class Loader(Protocol):
    """@public"""

    def __call__(self, filepath: str, content: str) -> LoaderResult:
        ...


# PORT COMMENT: No async at the moment
@runtime_checkable
class Transform(Protocol):
    """@public"""

    def __call__(
        self, cosmiconfig_result: Optional[CosmiconfigResult]
    ) -> Optional[CosmiconfigResult]:
        ...


class CommonOptions(BaseModel):
    """@public"""

    # PORT COMMENT: `config_prop` is our custom field, because in pyproject.toml
    #               we need to prefix the package name with `tool.`.
    #               Hence we must support multiple props to search for
    config_prop: Optional[Union[str, List[str]]] = Field(default=None)

    package_prop: Optional[Union[str, List[str]]] = Field(default=None)
    search_places: Optional[List[str]] = Field(default=None)
    ignore_empty_search_places: Optional[bool] = Field(default=None)
    stop_dir: Optional[str] = Field(default=None)
    cache: Optional[bool] = Field(default=None)


# PORT COMMENT: No async at the moment
class Options(CommonOptions):
    """@public"""

    loaders: Optional["Loaders"] = Field(default=None)
    transform: Optional[Transform] = Field(default=None)

    class Config:
        arbitrary_types_allowed = True


# PORT COMMENT: No async at the moment
class InternalOptions(BaseModel):
    """@internal"""

    apply_package_property_path_to_configuration: Optional[bool] = Field(default=None)
    meta_config_file_path: Optional[str] = Field(default=None)

    # PORT COMMENT: The below is equivalent to `Required<Options>`
    # PORT COMMENT 2: `config_prop` is our custom field.
    config_prop: Optional[Union[str, List[str]]] = Field(default=None)
    package_prop: Union[str, List[str]] = Field(default=None)
    search_places: List[str] = Field(default=None)
    ignore_empty_search_places: bool = Field(default=None)
    stop_dir: str = Field(default=None)
    cache: bool = Field(default=None)
    loaders: "Loaders" = Field(default=None)
    transform: Transform = Field(default=None)

    class Config:
        arbitrary_types_allowed = True


# PORT COMMENT: Defined separately from InternalOptions because
#               these values can be nullable.
class InternalOverrideOptions(Options):
    """@internal"""

    apply_package_property_path_to_configuration: Optional[bool] = Field(default=None)
    meta_config_file_path: Optional[str] = Field(default=None)


# PORT COMMENT: No async at the moment
Cache = Dict[str, Optional[CosmiconfigResult]]
"""@internal"""


# PORT COMMENT: No async at the moment
Loaders = Dict[str, Loader]
"""@public"""


class PublicExplorerBase(BaseModel):
    """@public"""

    clear_load_cache: Callable[[], None]
    clear_search_cache: Callable[[], None]
    clear_caches: Callable[[], None]


@runtime_checkable
class PublicExplorerSearchFn(Protocol):
    def __call__(self, search_from: Optional[str]) -> Optional[CosmiconfigResult]:
        ...


@runtime_checkable
class PublicExplorerLoadFn(Protocol):
    def __call__(self, filepath: str) -> Optional[CosmiconfigResult]:
        ...


# PORT COMMENT: No async at the moment
class PublicExplorer(PublicExplorerBase):
    """@public"""

    search: PublicExplorerSearchFn
    load: PublicExplorerLoadFn

    class Config:
        arbitrary_types_allowed = True



import importlib
from importlib.util import spec_from_file_location, module_from_spec
import json
from pathlib import Path

import toml
import yaml


# PORT COMMENT: We import from Py files instead of JS files,
#               and the imported Py file must expose a "config" variable.
def load_py(filepath: str, content: str):
    # Try to import file as module
    try:
        module = importlib.import_module(filepath)
        return module.config
    except ModuleNotFoundError:
        pass

    # Otherwise try to import file as python source file
    # See https://stackoverflow.com/a/54956419/9788634
    modulename = Path(filepath).stem
    spec = spec_from_file_location(modulename, filepath)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.config


def load_json(filepath: str, content: str):
    try:
        return json.loads(content)
    except json.JSONDecodeError as e:
        raise ValueError(f"JSON Error in {filepath}:\n{str(e)}")


def load_yaml(filepath: str, content: str):
    try:
        return yaml.load(content, Loader=yaml.FullLoader)
    except yaml.YAMLError as e:
        raise ValueError(f"YAML Error in {filepath}:\n{str(e)}")


def load_toml(filepath: str, content: str):
    try:
        return toml.loads(content)
    except toml.TomlDecodeError as e:
        raise ValueError(f"TOML Error in {filepath}:\n{str(e)}")
